#ifndef __FTPD_P_H__
#define __FTPD_P_H__ 1

#include "ftpd.h"
#include "log_unix.h"
#ifdef USE_PAM
# include "log_pam.h"
#endif
#ifdef WITH_LDAP
# include "log_ldap.h"
#endif
#ifdef WITH_MYSQL
# include "log_mysql.h"
#endif
#ifdef WITH_PUREDB
# include "log_puredb.h"
#endif
#if !defined(NO_GETOPT_LONG) && !defined(HAVE_GETOPT_LONG)
# include "gnu-getopt.h"
#else
# ifdef HAVE_GETOPT_H
#  include <getopt.h>
# endif
#endif

#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(st) ((unsigned) (st) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(st) (((st) & 0xff) == 0)
#endif

#define DEFAULT_BACKLOG 42
#define DEFAULT_BACKLOG_DATA 8
#define NICE_VALUE 10
#define THROTTLING_UNIT 10000UL
#define MAX_GROUPS 256
#define MAX_PASSWD_TRIES 5        /* Abort after 5 authentication failures */
#define PASSWD_FAILURE_DELAY (3UL*1000000UL)    /* Delay after each failure */
#define ASCII_CHUNKSIZE 65536U

#ifndef NO_STANDALONE
static volatile sig_atomic_t stop_server;
static const char *standalone_port = DEFAULT_FTP_PORT_S;
static const char *standalone_ip;
static volatile unsigned int nb_children;
static int listenfd = -1;
#endif

struct reply {
    struct reply *prev;
    char line[1];
};

static const char *GETOPT_OPTIONS =
    "14Aa:bc:"
#ifndef NO_STANDALONE
    "BC:"
#endif
    "dDeEf:"
#ifdef COOKIE
    "F:"
#endif
#ifndef NO_STANDALONE
    "g:"
#endif
    "hHiI:jk:Kl:L:Mm:N"
#ifdef QUOTAS
    "n:"
#endif
#if defined(WITH_UPLOAD_SCRIPT) && !defined(NON_ROOT_FTP)
    "o"
#endif
#ifdef WITH_ALTLOG
    "O:"
#endif
    "p:P:"
#ifdef RATIOS
    "q:Q:"
#endif
    "rRs"
#ifndef NO_STANDALONE
    "S:"
#endif
#ifdef THROTTLING
    "t:T:"
#endif
    "u:U:"
#ifdef VUSERS
    "v:"
#endif
    "V:wWxXz";

#ifndef NO_GETOPT_LONG    
static struct option long_options[] = {
    { "logpid", 0, NULL, '1' },
    { "ipv4only", 0, NULL, '4' },
    { "chrooteveryone", 0, NULL, 'A' },
    { "trustedgid", 1, NULL, 'a' },
    { "brokenclientscompatibility", 0, NULL, 'b' },
    { "maxclientsnumber", 1, NULL, 'c' },
# ifndef NO_STANDALONE
    { "daemonize", 0, NULL, 'B' },                
    { "maxclientsperip", 1, NULL, 'C' },
# endif
    { "verboselog", 0, NULL, 'd' },
    { "displaydotfiles", 0, NULL, 'D' },
    { "anonymousonly", 0, NULL, 'e' },
    { "noanonymous", 0, NULL, 'E' },
    { "syslogfacility", 1, NULL, 'f' },
# ifdef COOKIE
    { "fortunesfile", 1, NULL, 'F' },
# endif
# ifndef NO_STANDALONE
    { "pidfile", 1, NULL, 'g' },
# endif
    { "help", 0, NULL, 'h' },
    { "dontresolve", 0, NULL, 'H' },
    { "maxidletime", 1, NULL, 'I' },
    { "anonymouscantupload", 0, NULL, 'i' },
    { "createhomedir", 0, NULL, 'j' },
    { "keepallfiles", 0, NULL, 'K' },    
    { "maxdiskusagepct", 1, NULL, 'k' },
    { "login", 1, NULL, 'l' },
    { "limitrecursion", 1, NULL, 'L' },
    { "anonymouscancreatedirs", 0, NULL, 'M' },
    { "maxload", 1, NULL, 'm' },
    { "natmode", 0, NULL, 'N' },
# ifdef QUOTAS
    { "quota", 1, NULL, 'n' },
# endif
# if defined(WITH_UPLOAD_SCRIPT) && !defined(NON_ROOT_FTP)
    { "uploadscript", 0, NULL, 'o' },
# endif
# ifdef WITH_ALTLOG
    { "altlog", 1, NULL, 'O' },
# endif
    { "passiveportrange", 1, NULL, 'p' },
    { "forcepassiveip", 1, NULL, 'P' },
# ifdef RATIOS
    { "anonymousratio", 1, NULL, 'q' },
    { "userratio", 1, NULL, 'Q' },
# endif
    { "autorename", 0, NULL, 'r' },
    { "nochmod", 0, NULL, 'R' },    
    { "antiwarez", 0, NULL, 's' },
# ifndef NO_STANDALONE
    { "bind", 1, NULL, 'S' },
# endif
    { "anonymousbandwidth", 1, NULL, 't' },
    { "userbandwidth", 1, NULL, 'T' },
    { "umask", 1, NULL, 'U' },
    { "minuid", 1, NULL, 'u' },
# ifdef VUSERS
    { "virtualusers", 1, NULL, 'v' },
# endif
    { "trustedip", 1, NULL, 'V' },
    { "allowuserfxp", 0, NULL, 'w' },
    { "allowanonymousfxp", 0, NULL, 'W' },
    { "prohibitdotfileswrite", 0, NULL, 'x' },
    { "prohibitdotfilesread", 0, NULL, 'X' },
    { "allowdotfiles", 0, NULL, 'z' },        
    { NULL, 0, NULL, 0 }
};

#endif

#ifdef WITH_ALTLOG
static const AltLogPrefixes altlogprefixes[] = {
    { "clf", ALTLOG_CLF },
    { "stats", ALTLOG_STATS },
    { NULL, ALTLOG_NONE }
};

# define ALTLOG_DELIMITER ':'
# define ALTLOG_DEFAULT ALTLOG_CLF
#endif

#ifndef HAVE_SYS_FSUID_H
static sigset_t old_sigmask;
#endif

#ifndef HAVE_SETPROCTITLE
static char **argv0;
static int argv_lth;
#endif

/*
 * An authentication handler has three functions:
 * - One - parse() - is called with an optional file name, that contains a
 *   configuration file, or whatever is passed in the -l command-line switch 
 *   for this authentication.
 * - Another one, check() is called when the user has entered his password.
 *   It should fill an AuthResult structure.
 * - The last one - exit() - is called when the session is closed, and
 *   should free all internal allocated structures.
 */

typedef struct Authentication_ {
    const char * name;
    void (* parse)(const char * const file);
    void (* check)(AuthResult * const result,
                   const char *account, const char *password,
                   const struct sockaddr_storage * const sa,
                   const struct sockaddr_storage * const peer);
    void (* exit)(void);
} Authentication;

static Authentication auth_list[] = {
    { "unix", pw_unix_parse, pw_unix_check, pw_unix_exit },   /* 0 */
#ifdef USE_PAM
    { "pam", pw_pam_parse, pw_pam_check, pw_pam_exit },   /* 1 */
#endif
#ifdef WITH_MYSQL
    { "mysql", pw_mysql_parse, pw_mysql_check, pw_mysql_exit },   /* 2 */
#endif
#ifdef WITH_LDAP
    { "ldap", pw_ldap_parse, pw_ldap_check, pw_ldap_exit },   /* 3 */
#endif
#ifdef WITH_PUREDB
    { "puredb", pw_puredb_parse, pw_puredb_check, pw_puredb_exit },   /* 4 */
#endif
    { NULL, NULL, NULL, NULL }
};

#ifdef USE_PAM
# define DEFAULT_AUTHENTICATION (&auth_list[1])   /* pam */
#else
# define DEFAULT_AUTHENTICATION (&auth_list[0])   /* unix */
#endif

typedef struct Authentications_ {
    const Authentication *auth;
    char *conf_file;
    struct Authentications_ *next;
} Authentications;

static Authentications *first_authentications, *last_authentications;

#endif
